import numpy as np
import pandas as pd
import matplotlib.pyplot as plt
from scipy import stats, optimize
from sympy import Matrix, solve_linear_system, var
from IPython.display import display

from .utilities import prettyPrint

class OLS:
    """
    Example Use:
    df = pd.DataFrame({'y': [13, 15, 17, 18, 19, 22, 25, 27, 29, 30, 32, 35],
                   'x': [0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11]})
    y, x = df['y'], df['x']
    mod = OLS(y, x)
    res = mod.fit(const=True, cov_type='HC1')  # add constant and use heteroskedastic White variance estimate
    mod.summary(res)
    mod.plot_single_variable(res, xlab='x', ylab='y')
    """
    def __init__(self, Y, X, dropna=False):
        """
        Oridnary Least Squares Regression class
        :param Y: Dataframe column of dependent variable
        :param X: Series of explanatory variable
        """

        # Make X and Y Dataframes
        self.x = pd.DataFrame(X)

        if isinstance(Y, pd.DataFrame):
            Y.columns = ['y']
            self.y = Y
        elif isinstance(Y, pd.Series):
            Y = pd.DataFrame(Y)
            Y.columns = ['y']
            self.y = Y
        else:
            self.y = pd.DataFrame(Y, columns=['y'])

        if dropna:
            combined = self.x.join(self.y, how='inner')
            combined.dropna(inplace=True)
            self.y = combined['y']
            self.x = combined.drop('y', axis=1)

        self.preds = None  # predictions

    def fit(self, const=True, cov_type='HC1', ci=95):
        """
        Fit OLS by minimizing sum of squares
        :param const: Add constant to regression {Default: True}
        :param cov_type: Covariance type {Homoskedastic: 'hom', Heteroskedastic White: 'het',
                                          White biased: 'HC0', White unbiased: 'HC1'}
        :param ci: Confidence Interval expressed as integer percent {Default: 95 (%)}
        :return: Dictionary of results including test statistics
        """

        results = {}

        # Add 1's column if required
        if const:
            labels = ['const'] + list(self.x)
            self.x['const'] = 1
            self.x = self.x[labels]

        # Solve OLS equation B = (X'X)^-1 (X'Y)
        X = self.x.values
        Y = np.squeeze(self.y.values)

        n = X.shape[0]  # n observations
        k = X.shape[1]  # k predictors
        p = k - 1
        coeffs = np.squeeze(np.dot(np.linalg.inv(np.dot(X.T, X)), np.dot(X.T, Y)))
        results['coeffs'] = coeffs
        results['n'] = n
        results['df_resids'] = n - k
        results['df_model'] = p

        # Use model to evaluate residuals and MSE
        self.preds = np.dot(X, coeffs)
        results['preds'] = self.preds
        e = Y - self.preds
        mse = np.dot(e, e) / (n - k)
        results['resids'] = e
        results['mse'] = mse

        # Calculate R^2 and adj R^2
        et = Y - np.mean(Y)
        r2 = 1 - np.dot(e, e) / np.dot(et, et)
        adj_r2 = r2 - (1 - r2) * p / (n - p - 1)
        results['R^2'] = r2
        results['adj. R^2'] = adj_r2

        # F-test
        em = self.preds - np.mean(Y)
        msm = np.dot(em, em) / p
        F = msm / mse
        p_F = 1 - stats.f.cdf(F, p, n - p - 1)
        results['F'] = F
        results['p_F'] = p_F

        # Log-likelihood
        ssr = np.dot(e, e)  # sum square residuals
        s2 = ssr / n
        L = np.log((2 * np.pi * s2) ** (-n / 2) * np.exp(-ssr / (2 * s2)))
        results['L'] = L

        # AIC and BIC
        AIC = 2 * k - 2 * L
        BIC = k * np.log(n) - 2 * L
        results['AIC'] = AIC
        results['BIC'] = BIC

        # Variance-covariance matrix
        if cov_type in ['het', 'hetero', 'heteroskedastic', 'HC1']:  # Unbiased White estimator
            vcm = np.diag(
                np.linalg.inv(X.T.dot(X)).dot((X.T.dot(np.diag(e ** 2)).dot(X))).dot(np.linalg.inv(X.T.dot(X))))
            vcm = vcm * n / (n - k)  # make unbiased White estimator
            results['cov_type'] = 'HC1'
        elif cov_type == 'HC0':  # Biased White estimator
            vcm = np.diag(
                np.linalg.inv(X.T.dot(X)).dot((X.T.dot(np.diag(e ** 2)).dot(X))).dot(np.linalg.inv(X.T.dot(X))))
            results['cov_type'] = 'HC0'
        elif cov_type in ['hom', 'homo', 'homoskedastic']:
            vcm = np.diag(mse * np.linalg.inv(X.T.dot(X)))
            results['cov_type'] = 'Homoskedastic'
        else:
            raise ValueError("cov_type unknown, select either 'het' or 'hom'")
        results['VCM'] = vcm

        # Standard error
        se = np.sqrt(vcm)
        results['se'] = se

        # t-statistic & z-statistic
        tz = coeffs / se
        if cov_type in ['het', 'hetero', 'heteroskedastic', 'HC1', 'HC0']:  # White estimator
            # z-statistic
            test = 'z'
            p_z = 2 * stats.norm.sf(abs(tz))
            p_test = p_z
            results['z'] = tz
            results['p_z'] = p_z

        elif cov_type in ['hom', 'homo', 'homoskedastic']:
            # t-statistic
            test = 't'
            p_t = 2 * stats.t.sf(np.abs(tz), n - k)
            p_test = p_t
            results['t'] = tz
            results['p_t'] = p_t

        # Confidence Intervals
        ci_ = se * stats.t.ppf(1 - (1 - ci / 100) / 2, n - k)
        ci_left = (1 - ci / 100) / 2  # left tail value
        ci_right = 1 - (1 - ci / 100) / 2  # right tail value
        results['ci'] = ci_

        # Skewness and Kurtosis
        S = np.mean(e ** 3) / np.mean(e ** 2) ** (3 / 2)
        K = np.mean(e ** 4) / np.mean(e ** 2) ** 2
        results['skew'] = S
        results['kurtosis'] = K

        # Omnibus
        def Z1(S, N):
            Y = S * np.sqrt(((N + 1) * (N + 3)) / (6.0 * (N - 2.0)))
            b = 3.0 * (N ** 2.0 + 27.0 * N - 70) * (N + 1.0) * (N + 3.0)
            b /= (N - 2.0) * (N + 5.0) * (N + 7.0) * (N + 9.0)
            W2 = - 1.0 + np.sqrt(2.0 * (b - 1.0))
            alpha = np.sqrt(2.0 / (W2 - 1.0))
            z = 1.0 / np.sqrt(np.log(np.sqrt(W2)))
            z *= np.log(Y / alpha + np.sqrt((Y / alpha) ** 2.0 + 1.0))
            return z

        def Z2(K, N):
            E = 3.0 * (N - 1.0) / (N + 1.0)
            v = 24.0 * N * (N - 2.0) * (N - 3.0)
            v /= (N + 1.0) ** 2.0 * (N + 3.0) * (N + 5.0)
            X = (K - E) / np.sqrt(v)
            b = (6.0 * (N ** 2.0 - 5.0 * N + 2.0)) / ((N + 7.0) * (N + 9.0))
            b *= np.sqrt((6.0 * (N + 3.0) * (N + 5.0)) / (N * (N - 2.0) * (N - 3.0)))
            A = 6.0 + (8.0 / b) * (2.0 / b + np.sqrt(1.0 + 4.0 / b ** 2.0))
            z = (1.0 - 2.0 / A) / (1.0 + X * np.sqrt(2.0 / (A - 4.0)))
            z = (1.0 - 2.0 / (9.0 * A)) - z ** (1.0 / 3.0)
            z /= np.sqrt(2.0 / (9.0 * A))
            return z

        omni = Z1(S, n) ** 2 + Z2(K, n) ** 2
        p_omni = 1 - stats.chi2(2).cdf(omni)
        results['omnibus'] = omni
        results['p_omnibus'] = p_omni

        # Durbin-Watson Test
        DW = np.sum(np.diff(e) ** 2) / ssr
        results['DW'] = DW

        # Jarque-Bera Test
        JB = (n / 6) * (S ** 2 + (1 / 4) * (K - 3) ** 2)
        p_JB = 1 - stats.chi2(2).cdf(JB)
        results['JB'] = JB
        results['p_JB'] = p_JB

        # Condition Number
        X_mat = np.asmatrix(X)
        EV, _ = np.linalg.eig(X_mat.T * X_mat)
        CN = np.sqrt(np.max(EV) / np.min(EV))
        results['CN'] = CN

        # Summary table
        summary = pd.DataFrame({'coef': coeffs, 'std err': se, test: tz, 'P>|{}|'.format(test): p_test,
                                '[{:.3f}'.format(ci_left): coeffs - ci_,
                                '{:.3f}]'.format(ci_right): coeffs + ci_},
                               index=list(self.x))
        summary = summary[['coef', 'std err', test, 'P>|{}|'.format(test),
                           '[{:.3f}'.format(ci_left), '{:.3f}]'.format(ci_right)]]
        results['summary'] = summary.round(3)

        # Test table
        r = results
        table = pd.DataFrame(
            {'Statistic': ['Cov Type:', 'No. Obsv:', 'DF Resids:', 'DF Model:', 'R-squared:', 'Adj. R-squared:',
                           'MSE:', 'F-statistic:', 'P (F-statistic):', 'Log-Likelihood:'],
             'Values': [0, r['n'], r['df_resids'], r['df_model'], r['R^2'], r['adj. R^2'],
                        r['mse'], r['F'], r['p_F'], r['L']],
             'Statistic ': ['AIC:', 'BIC:', 'Omnibus:', 'P (Omnibus):', 'Skew:', 'Kurtosis:', 'Durbin-Watson:',
                            'Jarque-Bera:', 'P (JB):', 'Cond. No.:'],
             'Values ': [r['AIC'], r['BIC'], r['omnibus'], r['p_omnibus'], r['skew'], r['kurtosis'], r['DW'],
                         r['JB'], r['p_JB'], r['CN']]})
        table.set_index('Statistic', inplace=True)
        table = table.round(3)
        pd.options.mode.chained_assignment = None  # turn off warning for next line
        table['Values']['Cov Type:'] = r['cov_type']  # must be float for rounding, change 0 val to str
        table = table[['Values', 'Statistic ', 'Values ']]
        results['table'] = table

        return results

    def summary(self, results):
        """
        Print summary of results including all test statistics
        :param results: Dictionary of OLS.fit() results
        :return: print table of results to screen
        """

        prettyPrint(results['summary'])
        prettyPrint(results['table'])
    
    def jsummary(self, results):
        """
        Print summary of results including all test statistics in Jupyter Notebook
        :param results: dict; OLS.fit() results
        :return: display table of results in Jupyter Notebook
        """
    
        display(results['summary'])
        display(results['table'])

    def plot_single_variable(self, results, xlab=None, ylab=None, save=None):
        """
        Plot regression for single explanatory variable
        :param results: Dictionary of OLS.fit() results
        :param xlab: str x-axis label
        :param ylab: str y-axis label
        :param save: str of filename to save plot
        :return: plot of regression
        """

        n, m = self.x.shape
        if m != 2:
            raise ValueError('Error: can only plot single explanatory variable. Current X shape: ({},{})'.format(n, m))
        X = self.x.drop('const', axis=1)
        X = np.squeeze(X.values)
        Y = np.squeeze(self.y.values)
        plt.plot(X, Y, 'o', c='steelblue')
        Xmin = np.argmin(X)
        Xmax = np.argmax(X)
        plt.plot([X[Xmin], X[Xmax]], [self.preds[Xmin], self.preds[Xmax]], c='darkorange')
        if xlab:
            plt.xlabel(xlab)
        if ylab:
            plt.ylabel(ylab)
        if save:
            plt.savefig('{}.png'.format(save), dpi=200)
        plt.show()

class AR:
    
    def __init__(self, X, p=1):
        self.x = np.asarray(X)
        # create dataframe of lags
        self.lags = pd.DataFrame()
        for i in range(p):
            self.lags['phi_{}'.format(i)] = self.x[i:-p+i]
        self.lags['phi_{}'.format(p)] = self.x[p:]
        self.lags.dropna(inplace=True)
        
        self.y = np.array(self.lags['phi_0'])
        self.lags.drop(['phi_0'], axis=1, inplace=True)
    
    def fit(self, kendall_correction=False):
        """
            AR(p) OLS results of AR(1) model
            :param self.t0: ndarray; x_(t) values to run AR model on.
            :param self.lags: dataframe; columns of x_(t-1), x_(t-2), ..., x_(t-p) lags for AR model.
            :param kendall_correction: boolean; if True, perform Kendall correction on phi for phi ~ 1
            :return: Dictionary of OLS results
            """
        
        # Run OLS
        mod = eu.OLS(self.y, self.lags)
        res = mod.fit()
        
        # Kendall correction
        if kendall_correction:
            for i, phi in enumerate(res['coeffs'][1:]):  # remove constant coefficient
                if phi > 0.98:
                    phi_cor = phi + (1+3*phi) / len(self.y)
                    print('phi_{} after Kendall Correction: {:.5f}'.format(i + 1, phi_cor))
        else:
            for i, phi in enumerate(res['coeffs'][1:]):  # remove constant coefficient
                if phi > 0.98:
                    print('Warning: phi_{} = {:.5f}, may be persistent'.format(i + 1, phi))
        return res
    
    def summary(self, results, prec=3):
        """
        Print summary of results including all test statistics
        :param results: Dictionary of OLS.fit() results
        :return: print table of results to screen
        """
        
        prettyPrint(results['summary'])
        prettyPrint(results['table'])

    def jsummary(self, results):
        """
        Print summary of results including all test statistics in Jupyter Notebook
        :param results: dict; OLS.fit() results
        :return: display table of results in Jupyter Notebook
        """
            
        display(results['summary'])
        display(results['table'])

class ARCH_1:
    """
        ARCH(1) model with attributes fit() and predict()
        """
    def __init__(self, r):
        self.raw_ret = r
        self.ret = np.array(r.dropna())
        self.params = None
        self.err = None
    
    def fit(self, n):
        """
        Run ARCH(1) model n times, saving returning result with lowest standard error
        :param n: int; number of ARCH(1) models to run
        :return: (ndarray, ndarray); (parameter estimates, standard errors)
        for ARCH(1) model with minimum sum of standard errors
        """
        
        def log_likelihood_ARCH(theta, X):
            """
            Returns log likelihood of ARCH(1) model
            :param X: array: values to run ARCH model on
            :return: float: log likelihood
            """
            
            # Initialize variables
            x_t = X[1:-1]  # x(t)
            x_lag = X[:-2]  # x(t-1)
            
            # Compute residuals
            err = x_t - theta[0] - theta[1]*x_lag
            err_t = err[1:]
            err_lag = err[:-1]
            
            # Log likelihood function
            h_t = theta[2] + theta[3]*err_lag**2
            llf = 0.5 * np.sum(np.log(2*np.pi) + np.log(h_t) + err_t**2/h_t)
            return llf
        
        def single_run(ret):
            """
            Single ARCH Model run
            """
            theta0 = 0.001 + np.random.rand(4)
            bnds = ((0.001, 1), (0.001, 1), (0.001, 1), (0.001, 1))
            res = optimize.minimize(log_likelihood_ARCH, theta0, args=ret, bounds=bnds, method='L-BFGS-B')
            hess_inv = res.hess_inv.todense()
            se = np.array(np.sqrt(np.diag(hess_inv)))
            return res.x, se
        
        params = []
        err = []
        sum_err = np.zeros(n)
        for i in range(n):
            param_i, err_i = single_run(self.ret)
            params.append(param_i)
            err.append(err_i)
            sum_err[i] = np.sum(err_i)
        opt = int(np.argmin(sum_err))
        self.params = params[opt]
        self.err = err[opt]
        return params[opt], err[opt]
    
    def predict(self):
        # Initialize variables
        x_t = self.ret[1:-1]  # x(t)
        x_lag = self.ret[:-2]  # x(t-1)
        
        # Compute residuals
        err = x_t - self.params[0] - self.params[1]*x_lag
        err_lag = err[:-1]
        
        h_t = self.params[2] + self.params[3]*err_lag**2
        
        # Add results to original index
        preds = pd.Series(index=self.raw_ret.index)
        preds[-len(h_t) - 1:-1] = np.sqrt(h_t)
        return preds

class GARCH_11:
    """
    CH(1,1) model with attributes fit() and predict()
    """
    
    def __init__(self, r):
        self.raw_ret = r
        self.ret = np.array(r.dropna())
        self.params = None
        self.err = None
    
    def fit(self, n):
        """
        Run GARCH(1,1) model n times, saving returning result with lowest standard error
        :param n: int; number of GARCH(1,1) models to run
        :return: (ndarray, ndarray); (parameter estimates, standard errors)
        for GARCH(1,1) model with minimum sum of standard errors
        """
        
        def log_likelihood_GARCH(theta, X):
            """
            Returns log likelihood of ARCH(1) model
            :param X: array: values to run ARCH model on
            :return: float: log likelihood
            """
            
            # Initialize variables
            x_t = X[1:-1]  # x(t)
            x_lag = X[:-2]  # x(t-1)
            
            # Compute residuals
            err = x_t - theta[0] - theta[1]*x_lag
            err_t = err[1:]
            err_lag = err[:-1]
            
            h_t = np.zeros(len(err_t))
            h_t[0] = theta[2] + theta[3]*err_lag[0]**2
            for i in range(1, len(h_t)):
                h_t[i] = theta[2] + theta[3]*err_lag[i]**2 + theta[4]*h_t[i-1]
            
            # Log likelihood function
            llf = 0.5 * np.sum(np.log(2*np.pi) + np.log(h_t) + err_t**2/h_t)
            return llf
        
        def single_run(ret):
            """
            Single ARCH Model run
            """
            theta0 = 0.001 + np.random.rand(5)
            theta0[4] = 0.9
            bnds = ((0.001, 1), (0.001, 1), (0.0001, 1), (0.001, 1), (0.001, 1))
            res = optimize.minimize(log_likelihood_GARCH, theta0, args=ret, bounds=bnds, method='L-BFGS-B')
            hess_inv = res.hess_inv.todense()
            se = np.array(np.sqrt(np.diag(hess_inv)))
            return res.x, se
        
        params = []
        err = []
        sum_err = np.zeros(n)
        for i in range(n):
            param_i, err_i = single_run(self.ret)
            params.append(param_i)
            err.append(err_i)
            sum_err[i] = np.sum(err_i)
        opt = int(np.argmin(sum_err))
        self.params = params[opt]
        self.err = err[opt]
        return params[opt], err[opt]
    
    def predict(self):
        # Initialize variables
        x_t = self.ret[1:-1]  # x(t)
        x_lag = self.ret[:-2]  # x(t-1)
        
        # Compute residuals
        err = x_t - self.params[0] - self.params[1]*x_lag
        err_lag = err[:-1]
        
        h_t = self.params[2] + self.params[3]*err_lag**2
        h_t[0] = self.params[2] + self.params[3]*err_lag[0]**2
        for i in range(1, len(h_t)):
            h_t[i] = self.params[2] + self.params[3]*err_lag[i]**2 + self.params[4]*h_t[i-1]
        
        # Add results to original index
        preds = pd.Series(index=self.raw_ret.index)
        preds[-len(h_t) - 1:-1] = np.sqrt(h_t)
        return preds

def sym_solve_matrix(A, b):
    """
    Symbolically solve matrix in form Ax = b
    :param A: numpy matrix [N x M]
    :param b: numpy matrix [N x 1]
    :return: Dictonary of x_i values
    """

    N, M = A.shape
    v = var('x:{}'.format(M))
    system = Matrix(np.concatenate((A, b), axis=1))
    syms = 'v[0]'
    for i in range(M-1):
        syms = ''.join([syms, ', v[{}]'.format(i+1)])
    return eval('solve_linear_system(system, {})'.format(syms))


def t_test(a, b):
    s1 = np.std(a, ddof=1)
    s2 = np.std(b, ddof=1)
    n1 = len(a)
    n2 = len(b)
    t = (np.mean(a) - np.mean(b)) / np.sqrt(s1**2/n1 + s2**2/n2)
    dof = (s1**2/n1 + s2**2/n2)**2 / ((s1**2/n1)**2/(n1-1) + (s2**2/n2)**2/(n2-1))
    p = 2 * stats.t.sf(np.abs(t), dof)
    return t, p




import numpy as np
import pandas as pd
import pprint
import matplotlib.pyplot as plt
import seaborn as sns
import re

from matplotlib import cm
from tabulate import tabulate


class Memoize:
    def __init__(self, fn):
        self.fn = fn
        self.cache = {}

    def __call__(self, *args):
        if args not in self.cache:
            self.cache[args] = self.fn(*args)
        return self.cache[args]


def prettyPrint(obj):
    """Print object to screen in nice formatting"""
    if isinstance(obj, pd.DataFrame):
        print(tabulate(obj, headers='keys', tablefmt='psql'))
    elif isinstance(obj, dict):
        pp = pprint.PrettyPrinter(indent=4)
        pp.pprint(obj)
    else:
        raise ValueError('Error: {} not yet supported'.format(type(obj)))


def custom_sort(str_list, alphabet):
    """Custom sort of list of strings for given alphabet"""
    # Add all unique characters in list to end of provided alphabet.
    alphabet = alphabet + ''.join(list(set(''.join(str_list))))

    # Sort list with alphabet.
    return sorted(str_list, key=lambda x: [alphabet.index(c) for c in x])

def add_start(df, start_date):
    """
    Adds row of NaN values to dataframe with specified time index

    Parameters
    ----------
    df: pd.DataFrame
        input DataFrame
    start_date: str
        date to enter in dataframe

    Returns
    -------
    new_df: pd.DataFrame
        updated dataframe with empty row of added index
    """

    sd = pd.to_datetime(start_date)

    # Create time 0 dataframe row to insert
    t0 = {'time': sd}
    for col in list(df):
        t0[col] = None
    t0 = pd.DataFrame([t0], columns=t0.keys())
    t0.set_index('time', inplace=True)

    # Add t0 row to dataframe
    new_df = df.append(t0)
    new_df.sort_index(inplace=True)
    return new_df

def find_max_locs(df):
    """Find iloc location of maximum values in pd.DataFrame"""
    num_df = df.apply(pd.to_numeric, errors='coerce').values
    row_max = num_df.max(axis=1)
    ij_max = []

    for i, i_max in enumerate(row_max):
        if np.isnan(i_max):  # ignore nan values
            continue
        for j in range(num_df.shape[1]):
            if num_df[i, j] == i_max:
                ij_max.append((i, j))
    return(ij_max)

def find_min_locs(df):
    """Find iloc location of minimum values in pd.DataFrame"""
    num_df = df.apply(pd.to_numeric, errors='coerce').values
    row_min = num_df.min(axis=1)
    ij_min = []

    for i, i_min in enumerate(row_min):
        if np.isnan(i_min):  # ignore nan values
            continue
        for j in range(num_df.shape[1]):
            if num_df[i, j] == i_min:
                ij_min.append((i, j))
    return(ij_min)

def combine_errors_table(vals, errs, prec=3, latex_format=False):
    """
    Combines standard errors and values into DataFrame.

    Parameters
    ----------
    vals: pd.DataFrame
        values for table
    errs: pd.DataFrame
        errors for values
    latex_format: bool
        characters to use to separate values and errors
        {True: ' \pm ', False: ' +/- '}

    Returns
    -------
    df: pd.DataFrame
        DataFrame of values +/- errors
    """

    def build_error_column(val, err, latex_format):
        """Combine single column of values and errors with specifed separator"""
        v_e = []
        for i, (v, e) in enumerate(zip(val, err)):
            if latex_format:
                v_e.append(f'{v:.{prec}f} $\pm$ {abs(e):.{prec}f}')
            else:
                v_e.append(f'{v:.{prec}f} ± {abs(e):.{prec}f}')
        return np.asarray(v_e)

    if vals.size != errs.size:
        raise ValueError(f'vals size {vals.size} does not match errs {err.size}')
    df = pd.DataFrame(index=vals.index)

    for col in vals.columns:
        df[col] = build_error_column(vals[col], errs[col], latex_format)
    return df

def correlation(df, sort_col=None, plot=False):
    """
    Find correlation of DataFrame, plot results and/or
    return sorted correlations for a sincle column of
    the DataFrame.

    Parameters
    ----------
    df: pd.DataFrame
        DataFrame of input values
    sort_col: str, default=None
        column to sort correlations on. If provided,
        function returns DataFrame of results
    plot: bool, default=False
        if True plot correlation matrix

    Returns
    -------
    plot of correlation matrix if plot=True
    correlation results for one column if sort_col is True
    """

    corr = df.corr().abs()

    if plot:
        # Plot results
        fig, ax = plt.subplots(figsize=(12, 10))
        ax.matshow(corr)
        cmap = cm.get_cmap('coolwarm', 300)
        cax = ax.imshow(corr, interpolation="nearest", cmap=cmap)
        plt.xticks(range(len(corr.columns)), corr.columns)
        plt.yticks(range(len(corr.columns)), corr.columns)
        for tick in ax.get_xticklabels():
            tick.set_rotation(45)
        cax.set_clim([0, 1])
        fig.colorbar(cax, ticks=[0, .25, .5, .75, 1])


    if ret_col:
        return corr.sort_values([col], ascending=False)[col][1:]




def color_table(df, axis=1, title=None, rev_index=None, color='RdYlGn'):
    """
    Creates color coded comparison table from dataframe values (green high, red low)

    Parameters
    ----------
    df: pd.DataFrame
        DataFrame of values
    axis: int, default=1
        axis to normalize data upon
    title: str
        title for table
    rev_index: list(str)
        list of column names to reverse color coding
    color: str, default='RdYlGn'
        color palette for table

    Returns
    -------
        plot of color coded table
    """

    labels = df.values
    cdf = df.copy()
    if axis == 1:
        cdf -= cdf.mean(axis=0)
        cdf /= cdf.std(axis=0)
    else:
        cdf = cdf.transpose()
        cdf -= cdf.mean(axis=0)
        cdf /= cdf.std(axis=0)
        cdf = cdf.transpose()

    if rev_index:
        for i in rev_index:
            cdf.iloc[:, i] = 1 - cdf.iloc[:, i]

    plt.figure()
    if title:
        plt.title(title)
    sns.heatmap(cdf, cmap='RdYlGn', linewidths=0.5, annot=labels,
                fmt='0.1f', cbar=False)
    plt.xticks(rotation=0)
    plt.yticks(rotation=0)


def replace_multiple(text, dic):
    """replace multiple text fragments in a string"""
    for i, j in dic.items():
        text = text.replace(i, j)
    return text

def latex_print(obj,
                prec=3,
                matrix_type='b',
                col_fmt=None,
                caption=None,
                adjust=False,
                hide_index=False,
                int_vals=False,
                font_size=None,
                multi_row_header=False,
                bold_locs=None,
                ):

    r"""
    Returns object with syntax formatted for LaTeX

    Parameters
    ----------

    obj: {ndarray, matrix, list, pd.Series, pd.DataFrame}
        object to convert for printing in LaTeX
    prec: int
        precision for printing decimals

    Matrix Parameters
    -----------------
    matrix_type: str
        style for matrix in latex
        {'b': bracket, 'p': paranthesis}

    DataFrame Parameters
    --------------------
    col_fmt: str
        column format for latex, (e.g., 'lrc')
    caption: str
        add caption to table
    adjust: bool
        if True, adjust formatting to fit in LaTeX page
        width for large tables & arrays
         - requires \usepackage{adjustbox} in LaTeX preamble
    hide_index: bool
        if True remove index from printing in LaTeXint_vals: bool
        if True, remove decimal places if all are 0
        (e.g., 3.00 --> 3)
    font_size: str
        font size for tables
        {'tiny', 'scriptsize', 'footnotesize', 'small', 'normalsize',
        'large', 'Large', 'LARGE', 'huge', 'Huge'}
    multi_row_header: bool
        if True, use \thead to make header column multiple lines.
            - expects '*' as separator between lines in header column
            - requires \usepackage{makecell} in LaTeX preamble
    bold_locs: list(tuples)
        ilocs of table values to bold

    Returns
    -------
    prints LaTeX formatted object to screen
    """

    # Series
    if isinstance(obj, pd.Series):
        # Format to LaTeX list
        y = ['{:.{}f}'.format(elem, prec) for elem in list(obj)]
        y = ',\ '.join(y)
        y = '[' + y + ']'
        print(y)

    # DataFrame
    elif isinstance(obj, pd.DataFrame):
        # Format columns.
        if col_fmt:
            fmt = col_fmt
        else:
            fmt = 'l' + 'c'*len(list(obj))

        # Round values.
        obj = obj.round(prec)

        # Hide index if required
        if hide_index:
            obj.index = ['{}' for i in obj.index]

        # Remove NaN's.
        obj = obj.replace(np.nan, '-', regex=True)

        # Print table settings.
        print()
        print(r'﻿\begin{table}[H]')
        if font_size:
            print(r'\{}'.format(font_size))
        if caption:
            print(r'\caption{{{}}}'.format(caption))
        else:
            print(r'%\caption{}')
        print(r'\centering')
        if adjust:
            print(r'\begin{adjustbox}{width =\textwidth}')

        # Bold values if required.
        if bold_locs:
            for loc in bold_locs:
                obj.iloc[loc] = '\textbf{{{}}}'.format(obj.iloc[loc])

        # Replace improperly formatted values in table.
        reps = {}

        if int_vals:
            reps = {'.{} '.format('0'*i): '  '+' '*i for i in range(1, 10)}

        if multi_row_header:
            header = obj.to_latex().split('toprule\n')[1].split('\\\\\n\midrule')[0]
            new_header = ''
            for i, h in enumerate(header.split('&')):
                if i == 0:
                    new_header += '{}\n'
                elif i == len(header.split('&'))-1:
                    new_header += '& \\thead{{{}}}'.format(h).replace('*', '\\\\')
                else:
                    new_header += '& \\thead{{{}}}\n'.format(h).replace('*', '\\\\')
            reps[header] = new_header

        reps['\$'] = '$'
        reps[r'\textbackslash '] = '\\'
        reps['\{'] = '{'
        reps['\}'] = '}'
        reps['\_'] = '_'

        # Print table.
        print(replace_multiple(obj.to_latex(column_format=fmt), reps), end='')

        # Print end of table settings.
        if adjust:
            print(r'\end{adjustbox}')
        print(r'\end{table}')
        print()

    # Matrix
    elif isinstance(obj, np.matrix):
        if len(obj.shape) > 2:
            raise ValueError('matrix can at most display two dimensions')
        robj = obj.round(prec)
        lines = str(robj).replace('[', '').replace(']', '').splitlines()
        print()
        if matrix_type == 'b':
            rv = [r'\begin{bmatrix}']
            rv += ['  ' + ' & '.join(l.split()) + r'\\' for l in lines]
            rv += [r'\end{bmatrix}']
        elif matrix_type == 'p':
            rv = [r'\begin{pmatrix}']
            rv += ['  ' + ' & '.join(l.split()) + r'\\' for l in lines]
            rv += [r'\end{pmatrix}']
        else:
            raise ValueError("Error: matrix_type must be 'b' or 'p'")
        print('\n'.join(rv), '\n')
        print()

    elif isinstance(obj, np.ndarray):
        if len(obj.shape) == 1:
            y = ['{:.{}f}'.format(elem, prec) for elem in list(obj)]
            y = ',\ '.join(y)
            y = '[' + y + ']'
            print('\n', y, '\n')
        else:
            latex_print(np.matrix(obj))
    else:
        print('Error: datatype: {} is not defined in function latex_print')

def date_subset(df, start=None, end=None):
    """
    Subset DataFrame to given period

    Parameters
    ----------
    df: pd.DataFrame
        DataFrame of values
    start: {str, datetime obj, Timestamp}, default None
        start date, str fmt = '%m/%d/%Y'
    end: {str, datetime obj, Timestamp}, default None
        end date, str fmt = '%m/%d/%Y'

    Returns
    -------
    df: pd.DataFrame
        DataFrame subset to given dates
    """

    if start:
        df = df[df.index >= pd.to_datetime(start, infer_datetime_format=True)]
    if end:
        df = df[df.index <= pd.to_datetime(end, infer_datetime_format=True)]
    return df

def datetime_range(start, end, minute_delta):
    """
    Create list of dates with constant timedelta.

    Parameters
    ----------
    start: datetime obj
        starting date
    end datetime obj
        ending date
    minute_delta: int
        number of minutes for time delta

    Returns
    -------
    list of datetime objects
    """

    def dt_gen(start, end, delta):
        """Create generator of datetimes with given timedelta"""
        current = start
        while current < end:
            yield current
            current += delta

    return [dt for dt in dt_gen(start, end, timedelta(minutes=minute_delta))]

def remove_outer_periods(df, cols, starts, ends):
    """
    Replace all vales outside of specifed ranges
    in DataFrame columns with NaN's

    Parameters
    ----------
    df: pd.DataFrame
        DataFrame of values
    cols: list(str)
        column names
    starts: list(datetime obj)
        list of starting dates in col order
    ends: list[datetime obj]
        list of ending dates in col order

    Returns
    -------
    new_df: pd.DataFrame
        DataFrame with NaN's before and after specified dates
    """
    d = {}
    for c, s, e in zip(cols, starts, ends):
        df_c = df[c]
        df_c = pd.DataFrame(df_c[(df_c.index >= s) & (df_c.index <= e)])
        d[c] = df_c[~df_c.index.duplicated(keep='first')]  # drop dupes
    new_df = pd.DataFrame(index=df.index)
    new_df = new_df.join(d[c] for c in cols) # add each col

    return new_df

def plot_explained_variance(ev, type='bar', toff=0.02, boff=0.08, yoff=None, **kwargs):
    """
    Plot explained variance of PCA

    Parameters
    ----------
    ev: nd.array
        explained variance values
    type: str, default='bar'
        - 'bar': bar plots for cumulative variance
        - 'line': line plot for cumulative variance
    toff: float
        top x offset for cumulative variance values on plot
    boff: float
        bottom x offset for individual variance values on plot
    yoff: float
        tottom y offset for individual variance values on plot
    **kwargs:
        kwargs for plt.plot() if type==line for cumulative variance

    Returns
    -------
    plot of explained variance
    """

    ax = pd.Series(100*ev).plot(kind='bar', figsize=(10,5),
                   color='steelblue', fontsize=13)

    ax.set_ylabel('Explained Variance')
    ax.set_yticks([0, 20, 40, 60, 80, 100])
    ax.set_xlabel('Principal Component')

    # Create a list to collect the plt.patches data.
    totals = []

    # Find the values and append to list.
    for i in ax.patches:
        totals.append(i.get_height())

    # Set individual bar lables using patch list.
    total = sum(totals)

    yoff_d = {'bar': 0, 'line': 0.2}
    yoff_d[type] = yoff_d[type] if yoff is None else yoff

    # Set individual bar lables using patch list.
    for j, i in enumerate(ax.patches):
        # Get_x pulls left or right; get_height pushes up or down.
        if j > 0:
            ax.text(i.get_x()+boff, i.get_height()+1,
                    str(round(100*ev[j], 1))+'%', fontsize=11,
                    color='dimgrey')
        ax.text(i.get_x()+toff, 100*np.cumsum(ev)[j]+1+yoff_d[type],
                str(round(100*np.cumsum(ev)[j], 1))+'%', fontsize=12,
                color='dimgrey')

    xlab = [f'PC-{i+1}' for i, _ in enumerate(ax.patches)]
    if type == 'bar':
        ax2 = pd.Series(np.cumsum(100*ev)).plot(kind='bar', figsize=(10,5),
                   color='steelblue', fontsize=8, alpha=0.25)
        for i in ax2.patches:
            totals.append(i.get_height())
        ax2.set_xticklabels(xlab, rotation='horizontal')
        ax2.xaxis.grid(False)
    elif type == 'line':
        plt.plot(np.cumsum(100*ev), ':o', ms=8, c='steelblue', alpha=0.75, **kwargs)
        ax.xaxis.grid(False)
        ax.set_xticklabels(xlab, rotation='horizontal')



# __init__.py
from .utilities import prettyPrint, correlation, color_table, latex_print
from .utilities import date_subset, datetime_range, add_start
from .utilities import replace_multiple, plot_explained_variance, custom_sort
from .utilities import find_max_locs, combine_errors_table, remove_outer_periods
from .utilities import find_min_locs
from .empiricalmethods import OLS, AR, ARCH_1, sym_solve_matrix, t_test



from setuptools import setup, find_packages
from codecs import open
from os import path

here = path.abspath(path.dirname(__file__))

# Get the long description from the README file
with open(path.join(here, 'README.rst')) as f:
    long_description = f.read()

setup(
    name='empiricalutilities',
    version='0.1.3',
    description='A Python project for empirical data manipulation.',
    long_description=long_description,
    author='Jason R Becker',
    author_email='jasonrichardbecker@gmail.com',
    python_requires='>=3.6.0',
    license='MIT',
    classifiers=[
        'Development Status :: 3 - Alpha',
        'Intended Audience :: Developers',
        'Topic :: Software Development :: Build Tools',
        'License :: OSI Approved :: MIT License',
        'Programming Language :: Python :: 3.7',
        ],
    url='https://github.com/jason-r-becker/empiricalutilities',
    download_url='https://github.com/jason-r-becker/empiricalutilities/archive/0.1.tar.gz',
    keywords='empirical LaTeX OLS'.split(),
    packages=find_packages(exclude=['contrib', 'docs', 'tests', 'build']),
    install_requires=[
        'cycler',
        'kiwisolver',
        'matplotlib',
        'mpmath',
        'numpy',
        'pandas',
        'patsy',
        'pprint',
        'pyparsing',
        'pytz',
        'scipy',
        'seaborn',
        'six',
        'statsmodels',
        'sympy',
        'tabulate',
        ]
)
